from typing import Any, Dict, List, Optional, Union
from pydantic import BaseModel


class DatasetItem(BaseModel):
    uuid: Optional[str] = None
    inputs: Union[Dict[str, Any], List[Any], str]
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    metadata: Optional[Dict[str, Any]] = None


class Dataset(BaseModel):
    name: str
    created_at: str
    description: Optional[str] = None
    items: List[DatasetItem]




import os
import json
import asyncio
from typing import Dict, List, Optional, Union

from fastapi_poe import QueryRequest, PartialResponse, ProtocolMessage

async def main():  

    from weavel import create_poe_client, WeavelPoeClient

    client: WeavelPoeClient = create_poe_client()
    
    user_request = QueryRequest(
        version="1.1",
        type="query",
        query=[
            ProtocolMessage(
                role="system", content="you are a helpful assistant"
            ),
            ProtocolMessage(
                role="user",
                content="Hello 1",
                timestamp=1706514709229521,
            )
        ],
        conversation_id="c_test_1",
        user_id="u_test_1",
        message_id="m_test_1",
        metadata="",
    )
    
    assistant_responses: List[PartialResponse] = [
        PartialResponse(
            text="Hi, ",
        ),
        PartialResponse(
            text="How are you? 1",
        ),
    ]
    
    client.log(user_request, assistant_responses)
    
    user_request = QueryRequest(
        version="1.1",
        type="query",
        query=[
            ProtocolMessage(
                role="system", content="you are a helpful assistant"
            ),
            ProtocolMessage(
                role="user",
                content="how to use sympy library?",
                timestamp=1706514709229521,
            )
        ],
        conversation_id="c_test_2",
        user_id="u_test_2",
        message_id="m_test_2",
        metadata="",
    )
    
    assistant_responses: List[PartialResponse] = [
        PartialResponse(
            text="You can use ",
        ),
        PartialResponse(
            text="sympy by pip install sympy.",
        ),
    ]
    
    client.log(user_request, assistant_responses)

    client.close()
    
if __name__ == "__main__":
    asyncio.run(main())


from typing import Dict, Optional, Any, Literal
from datetime import datetime

from weavel._request import BaseModel


class Record(BaseModel):
    session_id: Optional[str] = None
    record_id: str
    created_at: Optional[datetime] = None
    metadata: Optional[Dict[str, str]] = None
    ref_record_id: Optional[str] = None


class Message(Record):
    type: Literal["message"] = "message"
    role: Literal["user", "assistant", "system"]
    content: str


class TrackEvent(Record):
    type: Literal["track_event"] = "track_event"
    name: str
    properties: Optional[Dict[str, str]] = None


class Trace(Record):
    type: Literal["trace"] = "trace"
    name: Optional[str] = None
    inputs: Optional[Dict[str, Any]] = None
    outputs: Optional[Dict[str, Any]] = None
    ended_at: Optional[datetime] = None




import os
import json
import asyncio

async def main():
    dataset = []
    with open("chat_alpaca/chatalpaca-1k-fixed.json", "r") as json_file:
        for line in json_file:
            dataset.append(json.loads(line))    

    from weavel import create_client

    client = create_client()

    sample_data = dataset[27]

    trace_uuid = client.start_trace(sample_data["user_uuid"], sample_data["created_at"])

    for message in sample_data["conversations"]:
        if message["role"] == "user":
            client.log.user_message(
                trace_uuid=trace_uuid,
                data_content=message["content"],
                timestamp=message["created_at"])
            print(message["content"])
        else:
            client.log.assistant_message(
                trace_uuid=trace_uuid,
                data_content=message["content"], 
                timestamp=message["created_at"]
            )
            print(message["content"])

    await asyncio.sleep(60)

    client.close()
    
if __name__ == "__main__":
    asyncio.run(main())


import os
from dotenv import load_dotenv

load_dotenv()

if os.getenv("WEAVEL_TESTING") == "true":
    BACKEND_SERVER_URL = "http://localhost:8000"
else:
    BACKEND_SERVER_URL = "https://api.weavel.ai"



from weavel import Weavel
from uuid import uuid4
from datetime import datetime
from dotenv import load_dotenv

load_dotenv()

start_time = datetime.now()
weavel = Weavel()
start_latency = (datetime.now() - start_time).total_seconds()
print("Start latency: ", start_latency)

latencies = []

user_identifier = "TEST USER ID 0630"

session_id='TEST SESSION ID 0630'
start_time = datetime.now()
session = weavel.session(user_id=user_identifier, session_id=session_id)
latencies.append((datetime.now() - start_time).total_seconds())

start_time = datetime.now()
session.track_event(
    name="paid",
    properties={"amount": "100"}
)
latencies.append((datetime.now() - start_time).total_seconds())

start_time = datetime.now()
session.track_event(
    name="test"
)
latencies.append((datetime.now() - start_time).total_seconds())

start_time = datetime.now()
session.message(
    role="user",
    content="Hi!",
)
latencies.append((datetime.now() - start_time).total_seconds())

start_time = datetime.now()
trace = session.trace(
    name="retrieval",
    record_id="test_trace_id"
)
latencies.append((datetime.now() - start_time).total_seconds())

start_time = datetime.now()
trace.log(
    name="DB-1"
)
latencies.append((datetime.now() - start_time).total_seconds())

start_time = datetime.now()
span = trace.span(
    name="search_engine"
)
latencies.append((datetime.now() - start_time).total_seconds())

start_time = datetime.now()
span.log(
    name="query",
    value="AAA"
)
latencies.append((datetime.now() - start_time).total_seconds())

start_time = datetime.now()
message = session.message(
    role="assistant",
    content="Hello!",
    metadata={"test_key": "test_value1"},
    reason_record_id=trace.record_id
)
latencies.append((datetime.now() - start_time).total_seconds())

for latency in latencies:
    print(latency)

# mean latency
print(sum(latencies) / len(latencies))

weavel.identify(
    user_id=user_identifier,
    properties={"name": "AAAA"}
)

start_time = datetime.now()
weavel.close()
final_latency = (datetime.now() - start_time).total_seconds()
print("Final latency: ", final_latency)



import os
import json
import asyncio

async def main():
    dataset = []
    with open("../chat_alpaca/retention_dataset.json", "r") as json_file:
        for line in json_file:
            dataset.append(json.loads(line))    

    from weavel import create_client

    client = create_client()

    for data in dataset: # trace 139 ~ 163
        print(data["id"])
        if data["trace_uuid"] is None:
            trace_uuid = client.start_trace(data["user_uuid"], data["created_at"])
        else:
            trace_uuid = data["trace_uuid"]

        for message in data["conversations"]:
            if message["role"] == "user":
                client.log.user_message(
                    trace_uuid=trace_uuid,
                    data_content=message["content"],
                    timestamp=message["created_at"])
            else:
                client.log.assistant_message(
                    trace_uuid=trace_uuid,
                    data_content=message["content"], 
                    timestamp=message["created_at"]
                )

    await asyncio.sleep(60)

    client.close()
    
if __name__ == "__main__":
    asyncio.run(main())


from typing import Dict, Optional
from datetime import datetime

from weavel._request import BaseModel


class Session(BaseModel):
    """Session object."""

    user_id: Optional[str]
    session_id: str
    created_at: datetime
    metadata: Optional[Dict[str, str]] = None



from __future__ import annotations

import time
from datetime import datetime, timezone
from typing import Any, Dict, List, Literal, Optional, Union
from threading import Thread
from concurrent.futures import Future, ThreadPoolExecutor

from weavel._request import (
    CaptureSessionRequest,
    IdentifyUserRequest,
    CaptureMessageRequest,
    CaptureTrackEventRequest,
    CaptureTraceRequest,
    CaptureSpanRequest,
    CaptureLogRequest,
    CaptureGenerationRequest,
    UpdateTraceRequest,
    UpdateSpanRequest,
    UpdateGenerationRequest,
    CaptureTestObservationRequest,
    UnionRequest,
)
from weavel._body import (
    CaptureSessionBody,
    CaptureMessageBody,
    CaptureTrackEventBody,
    CaptureTraceBody,
    CaptureSpanBody,
    CaptureLogBody,
    CaptureGenerationBody,
    UpdateTraceBody,
    UpdateSpanBody,
    UpdateGenerationBody,
    IdentifyUserBody,
    CaptureTestObservationBody,
)

from weavel._constants import BACKEND_SERVER_URL
from weavel._buffer_storage import BufferStorage
from weavel._api_client import APIClient, AsyncAPIClient
from weavel.utils import logger
from weavel.types.datasets import DatasetItem, Dataset


class Worker:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Worker, cls).__new__(cls)
        return cls._instance

    def __init__(
        self,
        api_key: str,
        base_url: Optional[str] = None,
        max_retry: int = 3,
        flush_interval: int = 60,
        flush_batch_size: int = 20,
    ) -> None:
        if not hasattr(self, "is_initialized"):
            self.api_key = api_key
            self.endpoint = BACKEND_SERVER_URL if not base_url else base_url
            self.endpoint += "/public/v2"
            self.max_retry = max_retry
            self.flush_interval = flush_interval
            self.flush_batch_size = flush_batch_size

            self.api_client = APIClient()
            self.async_api_client = AsyncAPIClient()

            self.api_pool = ThreadPoolExecutor(max_workers=1)
            self.buffer_storage = BufferStorage(max_buffer_size=1000)
            self._running = True
            self._thread = Thread(target=self.consume_buffer, daemon=True)
            self._thread.start()
            self.is_initialized = True
            self.testing = False

    def open_session(
        self,
        session_id: str,
        created_at: datetime,
        user_id: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> str:
        """Start the new session for user_id.

        Returns:
            The session id.
        """
        if not self.testing:
            if (
                created_at.tzinfo is None
                or created_at.tzinfo.utcoffset(created_at) is None
            ):
                created_at = created_at.astimezone(timezone.utc)
            else:
                created_at = created_at

            # add task to buffer
            request = CaptureSessionRequest(
                body=CaptureSessionBody(
                    user_id=user_id,
                    session_id=session_id,
                    created_at=created_at.isoformat(),
                    metadata=metadata,
                )
            )

            self.buffer_storage.push(request)

        return

    def identify_user(self, user_id: str, properties: Dict) -> None:
        """
        Identifies a user with the given properties.

        Args:
            user_id (str): The ID of the user.
            properties (Dict): The properties associated with the user.

        Returns:
            None
        """
        if not self.testing:
            request = IdentifyUserRequest(
                body=IdentifyUserBody(
                    user_id=user_id,
                    properties=properties,
                    created_at=datetime.now(timezone.utc).isoformat(),
                )
            )
            self.buffer_storage.push(request)

        return

    def capture_message(
        self,
        session_id: str,
        record_id: str,
        created_at: datetime,
        role: Literal["user", "assistant", "system"],
        content: str,
        metadata: Optional[Dict[str, Any]] = None,
        ref_record_id: Optional[str] = None,
    ):
        if not self.testing:
            if (
                created_at.tzinfo is None
                or created_at.tzinfo.utcoffset(created_at) is None
            ):
                created_at = created_at.astimezone(timezone.utc)
            else:
                created_at = created_at

            request = CaptureMessageRequest(
                body=CaptureMessageBody(
                    session_id=session_id,
                    record_id=record_id,
                    created_at=created_at.isoformat(),
                    role=role,
                    content=content,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_track_event(
        self,
        session_id: str,
        record_id: str,
        created_at: datetime,
        name: str,
        properties: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        ref_record_id: Optional[str] = None,
    ):
        if not self.testing:
            if (
                created_at.tzinfo is None
                or created_at.tzinfo.utcoffset(created_at) is None
            ):
                created_at = created_at.astimezone(timezone.utc)
            else:
                created_at = created_at

            request = CaptureTrackEventRequest(
                body=CaptureTrackEventBody(
                    session_id=session_id,
                    record_id=record_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    properties=properties,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_trace(
        self,
        session_id: str,
        record_id: str,
        created_at: datetime,
        name: str,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        ref_record_id: Optional[str] = None,
    ):
        if not self.testing:
            if (
                created_at.tzinfo is None
                or created_at.tzinfo.utcoffset(created_at) is None
            ):
                created_at = created_at.astimezone(timezone.utc)
            else:
                created_at = created_at

            request = CaptureTraceRequest(
                body=CaptureTraceBody(
                    session_id=session_id,
                    record_id=record_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def update_trace(
        self,
        record_id: str,
        ended_at: Optional[datetime] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        ref_record_id: Optional[str] = None,
    ):

        if not self.testing:
            request = UpdateTraceRequest(
                body=UpdateTraceBody(
                    record_id=record_id,
                    ended_at=ended_at.isoformat() if ended_at else None,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_span(
        self,
        record_id: str,
        observation_id: str,
        created_at: datetime,
        name: str,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
    ):
        if not self.testing:
            request = CaptureSpanRequest(
                body=CaptureSpanBody(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def update_span(
        self,
        observation_id: str,
        ended_at: Optional[datetime] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
    ):

        if not self.testing:
            request = UpdateSpanRequest(
                body=UpdateSpanBody(
                    observation_id=observation_id,
                    ended_at=ended_at.isoformat() if ended_at else None,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_log(
        self,
        record_id: str,
        observation_id: str,
        created_at: datetime,
        name: str,
        value: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
    ):
        if not self.testing:
            request = CaptureLogRequest(
                body=CaptureLogBody(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    value=value,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_generation(
        self,
        observation_id: str,
        created_at: datetime,
        name: str,
        record_id: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
        prompt_name: Optional[str] = None,
    ):
        if not self.testing:
            request = CaptureGenerationRequest(
                body=CaptureGenerationBody(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                    prompt_name=prompt_name,
                )
            )
            self.buffer_storage.push(request)
        return

    async def acapture_generation(
        self,
        observation_id: str,
        created_at: datetime,
        name: str,
        record_id: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
        prompt_name: Optional[str] = None,
    ):
        if not self.testing:
            request = CaptureGenerationRequest(
                body=CaptureGenerationBody(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                    prompt_name=prompt_name,
                )
            )
            self.buffer_storage.push(request)
        return

    def update_generation(
        self,
        observation_id: str,
        ended_at: Optional[datetime] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
        prompt_name: Optional[str] = None,
    ):
        if not self.testing:
            request = UpdateGenerationRequest(
                body=UpdateGenerationBody(
                    observation_id=observation_id,
                    ended_at=ended_at.isoformat() if ended_at else None,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_test_observation(
        self,
        created_at: datetime,
        name: str,
        test_uuid: str,
        dataset_item_uuid: str,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
    ):
        if self.testing:
            request = CaptureTestObservationRequest(
                body=CaptureTestObservationBody(
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    test_uuid=test_uuid,
                    dataset_item_uuid=dataset_item_uuid,
                )
            )
            self.buffer_storage.push(request)
        return

    def create_dataset(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> None:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            "/datasets",
            method="POST",
            json={
                "name": name,
                "description": description,
            },
        )
        if response.status_code != 200:
            raise Exception(f"Failed to create dataset: {response.text}")

    async def acreate_dataset(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> None:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            "/datasets",
            method="POST",
            json={
                "name": name,
                "description": description,
            },
        )
        if response.status_code != 200:
            raise Exception(f"Failed to create dataset: {response.text}")

    def create_dataset_items(
        self,
        dataset_name: str,
        items: List[DatasetItem],
    ) -> None:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            "/dataset_items/batch",
            method="POST",
            json={
                "dataset_name": dataset_name,
                "items": items,
            },
        )
        if response.status_code != 200:
            raise Exception(f"Failed to create dataset items: {response.text}")

    async def acreate_dataset_items(
        self,
        dataset_name: str,
        items: List[DatasetItem],
    ) -> None:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            "/dataset_items/batch",
            method="POST",
            json={
                "dataset_name": dataset_name,
                "items": items,
            },
        )
        if response.status_code != 200:
            raise Exception(f"Failed to create dataset items: {response.text}")

    def fetch_dataset(
        self,
        name: str,
    ) -> Dataset:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            f"/datasets/{name}",
            method="GET",
        )

        if response.status_code == 200:
            return Dataset(**response.json())
        else:
            raise Exception(f"Failed to get dataset: {response.text}")

    async def afetch_dataset(
        self,
        name: str,
    ) -> Dataset:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            f"/datasets/{name}",
            method="GET",
        )

        if response.status_code == 200:
            return Dataset(**response.json())
        else:
            raise Exception(f"Failed to get dataset: {response.text}")

    def create_test(
        self,
        test_uuid: str,
        dataset_name: str,
        tags: Optional[List[str]] = None,
    ) -> None:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            "/tests",
            method="POST",
            json={
                "test_uuid": test_uuid,
                "dataset_name": dataset_name,
                "tags": tags,
            },
        )
        if response.status_code != 200:
            raise Exception(f"Failed to create test: {response.text}")

    def send_requests(
        self,
        requests: List[UnionRequest],
    ):
        """
        Sends a batch of requests to the API endpoint.

        Args:
            requests (List[Union[OpenSessionBody, CaptureRecordBody, CaptureObservationBody]]):
                A list of requests to be sent to the API endpoint.

        Returns:
            None: This method does not return any value.

        Raises:
            Exception: If an error occurs while sending the requests.

        """
        # logger.info(requests)
        for attempt in range(self.max_retry):
            logger.info(requests)
            try:
                response = self.api_client.execute(
                    self.api_key,
                    self.endpoint,
                    "/batch",
                    method="POST",
                    json={"batch": [request.model_dump() for request in requests]},
                    timeout=10,
                )
                if response.status_code == 200:
                    return
                # from rich import print

                # print({"batch": [request.model_dump() for request in requests]})
            except Exception as e:
                print(e)
                time.sleep(2**attempt)
                continue

    def consume_buffer(self) -> None:
        """
        Continuously consumes the buffer and sends requests to the API.

        This method runs in a loop until the `_running` flag is set to False. It checks the buffer size and waits for
        the specified flush interval if the buffer is empty. If the buffer has data, it pulls a batch of data from
        the buffer and sends requests to the API using the `send_requests` method. The completion of the API task is
        handled by the `_handle_api_task_completion` method.

        Raises:
            Exception: If an error occurs while consuming the buffer.

        """
        while self._running:
            try:
                with self.buffer_storage.buffer_lock:
                    if self.buffer_storage.buffer_size == 0:
                        self.buffer_storage.not_empty_cv.wait(self.flush_interval)
                        continue
                    out = self.buffer_storage.pull(self.flush_batch_size)
                    future = self.api_pool.submit(self.send_requests, out)
                    future.add_done_callback(self._handle_api_task_completion)
                    self.buffer_storage.not_empty_cv.wait(self.flush_interval)
            except Exception as e:
                print(e)

    def flush(self) -> None:
        """
        Flushes the buffer by sending the buffered requests.

        This method pulls all the requests from the buffer and sends them in batches
        of size `self.flush_batch_size` to the `send_requests` method.

        Raises:
            Exception: If an error occurs while flushing the buffer.
        """
        try:
            with self.buffer_storage.buffer_lock:
                if self.buffer_storage.buffer_size == 0:
                    return
                out = self.buffer_storage.pull_all()
                # for request chunks size of self.flush_batch_size
                for i in range(0, len(out), self.flush_batch_size):
                    request = out[i : i + self.flush_batch_size]
                    self.send_requests(request)
        except Exception as e:
            print(e)

    def stop(self) -> None:
        """
        Stops the worker thread and flushes any remaining data in the buffer.

        This method sets the `_running` flag to False, notifies the buffer storage
        that it is not empty, joins the worker thread, and then calls the `flush`
        method to ensure any remaining data in the buffer is processed.

        Returns:
            None
        """
        self._running = False
        with self.buffer_storage.buffer_lock:
            self.buffer_storage.not_empty_cv.notify()
        self._thread.join()
        self.flush()

    def _handle_api_task_completion(self, future: Future) -> None:
        try:
            future.result()
        except Exception as e:
            print(e)



from weavel import Weavel
from uuid import uuid4
from datetime import datetime
from dotenv import load_dotenv

load_dotenv()

weavel = Weavel()

latencies = []

user_identifier = "TEST USER ID 0714"

session_id='TEST SESSION ID 0714'

# normal usage
session = weavel.session(user_id=user_identifier, session_id=session_id)

session.track(
    name="track_event_1",
    properties={"amount": "1 00"}
)

session.track(
    name="track_event_2"
)

session.message(
    role="user",
    content="Hi!",
)

session.trace(
    name="trace_1",
    inputs={"input_1": "input_1_value"},
    outputs={"output_1": "output_1_value"}
)

trace_id = "test_trace_id"
trace = session.trace(
    name="trace_2",
    record_id=trace_id
)

trace.span(
    name="span_1"
)

span = trace.span(
    name="span_2",
    inputs={"input_2": "input_2_value"},
    outputs={"output_2": "output_2_value"}
)

generation = span.generation(
    name="generation_1",
    inputs={"input_3": "input_3_value"},
    outputs={"output_3": "output_3_value"}
)

span_3 = span.span(
    name="span_3"
)

# update
span_3.update(
    inputs={"input_3": "input_3_value"},
    outputs={"output_3": "output_3_value"}
)
generation.update(
    inputs={"input_3": "input_3_new_value"},
)
trace.update(
    inputs={"input_4": "input_4_value"},
    outputs={"output_4": "output_4_value"}
)

# end
span_3.end()
generation.end()
span.end()
trace.end()

message = session.message(
    role="assistant",
    content="Hello!",
    metadata={"test_key": "test_value1"},
    ref_record_id=trace.record_id
)

# reuse session
session = weavel.session(session_id=session_id)

session.track(
    name="track_event_3",
    properties={"amount": "200"}
)

# reuse trace
trace = weavel.trace(
    record_id=trace_id
)

trace.log(
    name="log_1",
    value="log_1_value"
)

# create the new trace directly
trace_2 = weavel.trace(
    session_id=session_id,
    record_id="directly_created_trace_id",
    name="directly_created_trace",
)

trace_2.log(
    name="log_2",
    value="log_2_value"
)

previously_used_span_id = span.observation_id

# reuse span
span = weavel.span(
    observation_id=previously_used_span_id
)

span.log(
    name="log_3",
)

# create the new span directly
span_2 = weavel.span(
    record_id=trace.record_id,
    name="directly_created_span_2",
)

span_2.log(
    name="log_4",
)

# track directly

weavel.track(
    session_id=session_id,
    name="directly_created_track_1",
    properties={"amount": "300"}
)

weavel.close()



from typing import Dict, List, Optional, Any, Literal, Union
from uuid import uuid4
from datetime import datetime, timezone
from pydantic import BaseModel, Field, PrivateAttr

from typing import Literal
from weavel._worker import Worker
from .types import Session, Message, Trace, TrackEvent, Log, Span, Generation


class BaseClient(BaseModel):
    _weavel_client: Worker = PrivateAttr()

    def __init__(self, weavel_client: Worker, **data):
        super().__init__(**data)
        self._weavel_client = weavel_client


class ObjectClient(BaseClient):
    record_id: Optional[str] = None

    def log(
        self,
        name: str,
        value: Optional[str] = None,
        observation_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> Log:
        if observation_id is None:
            observation_id = str(uuid4())
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        log = Log(
            record_id=self.record_id,
            observation_id=observation_id,
            created_at=created_at,
            metadata=metadata,
            name=name,
            value=value,
        )
        self._weavel_client.capture_log(
            **{
                k: v
                for k, v in log.model_dump().items()
                if k != "type" and v is not None
            }
        )
        return log

    def generation(
        self,
        name: str,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        observation_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> "GenerationClient":
        if observation_id is None:
            observation_id = str(uuid4())
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        generation = Generation(
            record_id=self.record_id,
            observation_id=observation_id,
            created_at=created_at,
            name=name,
            inputs=inputs,
            outputs=outputs,
            metadata=metadata,
            parent_observation_id=(
                self.observation_id if isinstance(self, SpanClient) else None
            ),
        )
        generation_client = GenerationClient(
            **generation.model_dump(),
            weavel_client=self._weavel_client,
        )
        self._weavel_client.capture_generation(
            **{
                k: v
                for k, v in generation.model_dump().items()
                if k != "type" and v is not None
            }
        )

        return generation_client

    def span(
        self,
        name: str,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        observation_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> "SpanClient":
        if observation_id is None:
            observation_id = str(uuid4())
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        span = Span(
            record_id=self.record_id,
            observation_id=observation_id,
            created_at=created_at,
            name=name,
            inputs=inputs,
            outputs=outputs,
            metadata=metadata,
            parent_observation_id=(
                self.observation_id if isinstance(self, SpanClient) else None
            ),
        )
        span_client = SpanClient(
            **span.model_dump(),
            weavel_client=self._weavel_client,
        )
        self._weavel_client.capture_span(
            **{
                k: v
                for k, v in span.model_dump().items()
                if k != "type" and v is not None
            }
        )

        return span_client


class SessionClient(Session, BaseClient):
    def message(
        self,
        role: Literal["user", "assistant", "system"],
        content: str,
        record_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
        ref_record_id: Optional[str] = None,
    ) -> Message:
        if record_id is None:
            record_id = str(uuid4())
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        message = Message(
            session_id=self.session_id,
            record_id=record_id,
            created_at=created_at,
            metadata=metadata,
            ref_record_id=ref_record_id,
            role=role,
            content=content,
        )
        self._weavel_client.capture_message(
            **{
                k: v
                for k, v in message.model_dump().items()
                if k != "type" and v is not None
            }
        )
        return message

    def track(
        self,
        name: str,
        properties: Optional[Dict[str, str]] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
        record_id: Optional[str] = None,
        ref_record_id: Optional[str] = None,
    ) -> TrackEvent:
        if record_id is None:
            record_id = str(uuid4())
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        track_event = TrackEvent(
            session_id=self.session_id,
            record_id=record_id,
            created_at=created_at,
            name=name,
            properties=properties,
            metadata=metadata,
            ref_record_id=ref_record_id,
        )
        self._weavel_client.capture_track_event(
            **{
                k: v
                for k, v in track_event.model_dump().items()
                if k != "type" and v is not None
            }
        )
        return track_event

    def trace(
        self,
        name: str,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        record_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
        ref_record_id: Optional[str] = None,
    ) -> "TraceClient":
        if record_id is None:
            record_id = str(uuid4())
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        trace = Trace(
            session_id=self.session_id,
            record_id=record_id,
            created_at=created_at,
            name=name,
            inputs=inputs,
            outputs=outputs,
            metadata=metadata,
            ref_record_id=ref_record_id,
        )
        trace_client = TraceClient(
            **trace.model_dump(),
            weavel_client=self._weavel_client,
        )
        self._weavel_client.capture_trace(
            **{
                k: v
                for k, v in trace.model_dump().items()
                if k != "type" and v is not None
            }
        )
        return trace_client


class TraceClient(Trace, ObjectClient):

    def update(
        self,
        ended_at: Optional[datetime] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        ref_record_id: Optional[str] = None,
    ) -> None:
        if ended_at is None:
            ended_at = datetime.now(timezone.utc)

        self._weavel_client.update_trace(
            record_id=self.record_id,
            ended_at=ended_at,
            inputs=inputs,
            outputs=outputs,
            metadata=metadata,
            ref_record_id=ref_record_id,
        )
        self.ended_at = ended_at
        return

    def end(
        self,
        ended_at: Optional[datetime] = None,
    ) -> None:
        if ended_at is None:
            ended_at = datetime.now(timezone.utc)

        self._weavel_client.update_trace(
            record_id=self.record_id,
            ended_at=ended_at,
        )
        self.ended_at = ended_at
        return


class SpanClient(Span, ObjectClient):

    def update(
        self,
        ended_at: Optional[datetime] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
    ) -> None:
        if ended_at is None:
            ended_at = datetime.now(timezone.utc)

        self._weavel_client.update_span(
            observation_id=self.observation_id,
            ended_at=ended_at,
            inputs=inputs,
            outputs=outputs,
            metadata=metadata,
            parent_observation_id=parent_observation_id,
        )
        self.ended_at = ended_at
        return

    def end(
        self,
        ended_at: Optional[datetime] = None,
    ) -> None:
        if ended_at is None:
            ended_at = datetime.now(timezone.utc)

        self._weavel_client.update_span(
            observation_id=self.observation_id,
            ended_at=ended_at,
        )
        self.ended_at = ended_at
        return


class GenerationClient(Generation, ObjectClient):

    def update(
        self,
        ended_at: Optional[datetime] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
    ) -> None:
        if ended_at is None:
            ended_at = datetime.now(timezone.utc)

        self._weavel_client.update_generation(
            observation_id=self.observation_id,
            ended_at=ended_at,
            inputs=inputs,
            outputs=outputs,
            metadata=metadata,
            parent_observation_id=parent_observation_id,
        )
        self.ended_at = ended_at
        return

    def end(
        self,
        ended_at: Optional[datetime] = None,
    ) -> None:
        if ended_at is None:
            ended_at = datetime.now(timezone.utc)

        self._weavel_client.update_generation(
            observation_id=self.observation_id,
            ended_at=ended_at,
        )
        self.ended_at = ended_at
        return



import os
import json

# count the number of files in the directory daily_dialog_result
dd_dataset = []
dd_count = 0
for root, dirs, files in os.walk("daily_dialog_result"):
    for file in files:
        if file.endswith(".json"):
            dd_count += 1
            with open(os.path.join(root, file), "r") as f:
                data = json.load(f)
                dd_dataset.append(data)

# count the number of files in the directory chat_alpaca_result
ca_dataset = []
ca_count = 0
for root, dirs, files in os.walk("chat_alpaca_result"):
    for file in files:
        if file.endswith(".json"):
            ca_count += 1
            with open(os.path.join(root, file), "r") as f:
                data = json.load(f)
                ca_dataset.append(data)
                
import supabase
from dotenv import load_dotenv
load_dotenv()

url = os.getenv("SUPABASE_URL")
key = os.getenv("SUPABASE_KEY")

client = supabase.create_client(url, key)

# for ca in ca_dataset:
#     for trace_data in ca["trace_data"]:
#         uuid = trace_data["uuid"]
#         question_topics = trace_data["question_topics"]
#         information_topics = trace_data["information_topics"]
#         for topic in question_topics:
#             if topic["value"] is None:
#                 topic["value"] = ""
#         for topic in information_topics:
#             if topic["value"] is None:
#                 topic["value"] = ""

#         (
#             client.table("conversation_analysis")
#             .insert(
#                 {
#                     "trace_data_uuid": uuid,
#                     "question_topics": question_topics,
#                     "information_topics": information_topics,
#                 }
#             )
#             .execute()
#         )

project_uuid = "b7c360b6-a862-4ff3-ad31-d8a620a201dc"

# make unit
unit = (
    client.table("unit")
    .insert(
        {
            "name": "daily_dialog",
            "description": "daily_dialog dataset",
            "project_uuid": project_uuid,
        }
    )
    .execute()
).data

unit_uuid = unit[0]["uuid"]

for i, dd in enumerate(dd_dataset):
    # save trace_data
    # make user
    
    user = (
        client.table("user")
        .insert(
            {
                "name": f"daily dialog User {i}",
                "project_uuid": project_uuid,
            }
        )
        .execute()
    ).data
    user_uuid = user[0]["uuid"]
    
    # make trace
    
    trace = (
        client.table("trace")
        .insert(
            {
                "user_uuid": user_uuid,
                "project_uuid": project_uuid,
                "created_at": dd["dialog"][0]["created_at"],
                "updated_at": dd["dialog"][-1]["created_at"],
            }
        )
        .execute()
    ).data
    
    trace_uuid = trace[0]["uuid"]
    
    for trace_data in dd["dialog"]:
        trace_data_in_db = (
            client.table("trace_data")
            .insert(
                {
                    "trace_uuid": trace_uuid,
                    "unit_uuid": unit_uuid,
                    "type": trace_data["type"],
                    "content": trace_data["content"],
                    "created_at": trace_data["created_at"],
                    "metadata": {'source': 'daily_dialog'},
                    "project_uuid": project_uuid,
                }   
            )
            .execute()
        ).data
        uuid = trace_data_in_db[0]["uuid"]
        
        question_topics = trace_data["question_topics"]
        information_topics = trace_data["information_topics"]
        for topic in question_topics:
            if topic["value"] is None:
                topic["value"] = ""
        for topic in information_topics:
            if topic["value"] is None:
                topic["value"] = ""

        (
            client.table("conversation_analysis")
            .insert(
                {
                    "trace_data_uuid": uuid,
                    "question_topics": question_topics,
                    "information_topics": information_topics,
                }
            )
        )


from __future__ import annotations

import asyncio
from concurrent.futures import ThreadPoolExecutor
from itertools import islice
import os
from datetime import datetime, timezone
import time
from typing import Callable, Dict, List, Optional, Any, Union
from uuid import uuid4

from dotenv import load_dotenv
from weavel._worker import Worker

# from weavel.types.instances import Session, Span, Trace
from weavel.object_clients import (
    GenerationClient,
    SessionClient,
    SpanClient,
    TraceClient,
)
from weavel.types.datasets import Dataset, DatasetItem

load_dotenv()


class Weavel:
    """Client for interacting with the Weavel service.

    This class provides methods for creating and managing traces, tracking user actions,
    and closing the client connection.

    Args:
        api_key (str, optional): The API key for authenticating with the Weavel service.
            If not provided, the API key will be retrieved from the environment variable
            WEAVEL_API_KEY.

    Attributes:
        api_key (str): The API key used for authentication.

    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        max_retry: Optional[int] = 3,
        flush_interval: Optional[int] = 60,
        flush_batch_size: Optional[int] = 20,
    ):
        self.api_key = api_key or os.getenv("WEAVEL_API_KEY")
        assert self.api_key is not None, "API key not provided."
        self._worker = Worker(
            self.api_key,
            base_url=base_url,
            max_retry=max_retry,
            flush_interval=flush_interval,
            flush_batch_size=flush_batch_size,
        )

        self.testing = False

    def session(
        self,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> "SessionClient":
        """Create a new session for the specified user.

        Args:
            user_id (str): The user ID for the session.
            session_id (str, optional): The session ID. If not provided, a new session ID will be generated.
            created_at (datetime, optional): The created_at for the session. If not provided, the current time will be used.
            metadata (Dict[str, str], optional): Additional metadata for the session.

        Returns:
            SessionClient: The session object.

        """
        # if user_id is None and session_id is None:
        #     raise ValueError("user_id or session_id must be provided.")

        if session_id is None:
            session_id = str(uuid4())
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        session = SessionClient(
            user_id=user_id,
            session_id=session_id,
            created_at=created_at,
            metadata=metadata,
            weavel_client=self._worker,
        )

        if self.testing:
            return session

        # if user_id is not None:
        self._worker.open_session(
            session_id=session_id,
            created_at=created_at,
            user_id=user_id,
            metadata=metadata,
        )
        return session

    def identify(self, user_id: str, properties: Dict[str, Any]):
        """Identify a user with the specified properties.

        You can save any user information you know about a user for user_id, such as their email, name, or phone number in dict format.
        Properties will be updated for every call.
        """
        if self.testing:
            return
        self._worker.identify_user(user_id, properties)

    def trace(
        self,
        session_id: Optional[str] = None,
        record_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        name: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        ref_record_id: Optional[str] = None,
    ) -> TraceClient:
        """Create a new trace record or fetch an existing one

        Args:
            session_id (str, optional): The session ID for the trace.
            record_id (str, optional): The record ID. If not provided, a new record ID will be generated.
            created_at (datetime, optional): The created_at for the trace. If not provided, the current time will be used.
            name (str, optional): The name of the trace.
            inputs (Dict[str, Any], optional): The inputs for the trace.
            outputs (Dict[str, Any], optional): The outputs for the trace.
            metadata (Dict[str, Any], optional): Additional metadata for the trace.
            ref_record_id (str, optional): The record ID to reference.

        """

        if session_id is None and record_id is None:
            raise ValueError("session_id or record_id must be provided.")

        if session_id is not None and name is None:
            raise ValueError(
                "If you want to create a new trace, you must provide a name."
            )

        if session_id is not None:
            # Create a new trace
            if record_id is None:
                record_id = str(uuid4())
            if created_at is None:
                created_at = datetime.now(timezone.utc)

            if not self.testing:
                self._worker.capture_trace(
                    session_id=session_id,
                    record_id=record_id,
                    created_at=created_at,
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            return TraceClient(
                session_id=session_id,
                record_id=record_id,
                created_at=created_at,
                name=name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                weavel_client=self._worker,
            )
        else:
            # Fetch an existing trace
            return TraceClient(record_id=record_id, weavel_client=self._worker)

    def generation(
        self,
        record_id: Optional[str] = None,
        observation_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        name: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        parent_observation_id: Optional[str] = None,
        prompt_name: Optional[str] = None,
    ) -> GenerationClient:
        # if record_id, observation_id, parent_observation_id
        # None, None, None -> ValueError
        # None, Value -> Fetch
        # Value -> Create
        # None, -, Value -> Create

        if not record_id and not observation_id and not parent_observation_id:
            if observation_id is None:
                observation_id = str(uuid4())
            if created_at is None:
                created_at = datetime.now(timezone.utc)
            self._worker.capture_generation(
                record_id=record_id,
                observation_id=observation_id,
                created_at=created_at,
                name=name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                parent_observation_id=parent_observation_id,
                prompt_name=prompt_name,
            )
        if (
            record_id or (record_id is None and parent_observation_id)
        ) and name is None:
            raise ValueError(
                "If you want to create a new generation, you must provide a name."
            )

        if record_id is not None or (
            record_id is None and parent_observation_id is not None
        ):
            # Create a new generation
            if observation_id is None:
                observation_id = str(uuid4())
            if created_at is None:
                created_at = datetime.now(timezone.utc)

            if not self.testing:
                self._worker.capture_generation(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at,
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                    prompt_name=prompt_name,
                )
            return GenerationClient(
                record_id=record_id,
                observation_id=observation_id,
                created_at=created_at,
                name=name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                parent_observation_id=parent_observation_id,
                prompt_name=prompt_name,
                weavel_client=self._worker,
            )
        else:
            # Fetch an existing generation
            return GenerationClient(
                observation_id=observation_id,
                weavel_client=self._worker,
            )

    def span(
        self,
        record_id: Optional[str] = None,
        observation_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        name: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        parent_observation_id: Optional[str] = None,
    ) -> SpanClient:
        # if record_id, observation_id, parent_observation_id
        # None, None, None -> ValueError
        # None, Value -> Fetch
        # Value -> Create
        # None, -, Value -> Create

        if not record_id and not observation_id and not parent_observation_id:
            raise ValueError(
                "One of the record_id, observation_id, or parent_observation_id must be provided."
            )

        if (
            record_id or (record_id is None and parent_observation_id)
        ) and name is None:
            raise ValueError(
                "If you want to create a new span, you must provide a name."
            )

        if record_id is not None or (
            record_id is None and parent_observation_id is not None
        ):
            # Create a new span
            if observation_id is None:
                observation_id = str(uuid4())
            if created_at is None:
                created_at = datetime.now(timezone.utc)

            if not self.testing:
                self._worker.capture_span(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at,
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            return SpanClient(
                record_id=record_id,
                observation_id=observation_id,
                created_at=created_at,
                name=name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                parent_observation_id=parent_observation_id,
                weavel_client=self._worker,
            )
        else:
            # Fetch an existing span
            return SpanClient(observation_id=observation_id, weavel_client=self._worker)

    def track(
        self,
        session_id: str,
        name: str,
        properties: Optional[Dict[str, Any]] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
        record_id: Optional[str] = None,
        ref_record_id: Optional[str] = None,
    ) -> None:
        """Track an event for the specified session.

        Args:
            session_id (str): The session ID for the event.
            name (str): The name of the event.
            properties (Dict[str, Any], optional): Additional properties for the event.

        """
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        if not self.testing:
            self._worker.capture_track_event(
                session_id=session_id,
                record_id=record_id,
                created_at=created_at,
                name=name,
                properties=properties,
                metadata=metadata,
                ref_record_id=ref_record_id,
            )

    def create_dataset(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> None:
        """Upload a dataset to the Weavel service.

        Args:
            name (str): The name of the dataset.
            description (str): The description of the dataset.
        """
        if self.testing:
            return

        self._worker.create_dataset(
            name=name,
            description=description,
        )

    async def acreate_dataset(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> None:
        """Upload a dataset to the Weavel service.

        Args:
            name (str): The name of the dataset.
            description (str): The description of the dataset.
        """
        if self.testing:
            return

        await self._worker.acreate_dataset(
            name=name,
            description=description,
        )

    def get_dataset(self, name: str) -> Dataset:
        """
        Retrieves a dataset with the given name.

        Args:
            name (str): The name of the dataset to retrieve.

        Returns:
            Dataset: The retrieved dataset.
        """
        if self.testing:
            return {}

        return self._worker.fetch_dataset(name)

    async def aget_dataset(self, name: str) -> Dataset:
        """
        Retrieves a dataset with the given name.

        Args:
            name (str): The name of the dataset to retrieve.

        Returns:
            Dataset: The retrieved dataset.
        """
        if self.testing:
            return {}

        return await self._worker.afetch_dataset(name)

    def create_dataset_items(
        self,
        dataset_name: str,
        items: Union[List[Dict[str, Any]], List[DatasetItem]],
    ) -> None:
        """Upload dataset items to the Weavel service.

        Args:
            dataset_name (str): The name of the dataset.
            items (Union[List[Dict[str, Any]], List[DatasetItem]]): The dataset items to upload.
        """
        if self.testing:
            return

        for item in items:
            if not isinstance(item, DatasetItem):
                item = DatasetItem(**item)

        self._worker.create_dataset_items(dataset_name, items)

    async def acreate_dataset_items(
        self,
        dataset_name: str,
        items: Union[List[Dict[str, Any]], List[DatasetItem]],
    ) -> None:
        """Upload dataset items to the Weavel service.

        Args:
            dataset_name (str): The name of the dataset.
            items (Union[List[Dict[str, Any]], List[DatasetItem]]): The dataset items to upload.
        """
        if self.testing:
            return

        for item in items:
            if not isinstance(item, DatasetItem):
                item = DatasetItem(**item)

        await self._worker.acreate_dataset_items(dataset_name, items)

    def test(
        self,
        func: Callable,
        dataset_name: str,
        batch_size: int = 50,
        delay: int = 10,
        tags: Optional[List[str]] = None,
    ) -> None:
        """Test the function with the dataset.
        It must be used in the jupyter notebook environment.

        Args:
            func (callable): The function to test.
            dataset_name (str): The name of the dataset.
            tags: (List[str], optional): The tags for the test.
            batch_size (int, optional): The batch size for the test. Default is 50.
            delay (int, optional): The delay (in seconds) between each batch. Default is 10.
            mock_inputs (Dict[str, Any], optional): The mock input variables to run the function.
        """
        import nest_asyncio

        nest_asyncio.apply()

        # fetch dataset from server
        dataset: Dataset = self._worker.fetch_dataset(dataset_name)

        # create test instance in database
        test_uuid = str(uuid4())
        self._worker.create_test(test_uuid, dataset_name, tags)

        self.testing = True
        self._worker.testing = True

        def _runner(
            func: callable,
            inputs: Union[Dict[str, Any], List[Any], str],
            dataset_item_uuid: str,
            test_uuid: str,
        ):
            # run the function and capture the result
            if isinstance(inputs, str):
                result = func(inputs)
            else:
                result = func(**inputs)

            if not isinstance(result, dict) and not isinstance(result, str):
                result = str(result)

            self._worker.capture_test_observation(
                created_at=datetime.now(timezone.utc),
                name=dataset_name,
                test_uuid=test_uuid,
                dataset_item_uuid=dataset_item_uuid,
                inputs=inputs,
                outputs=result,
            )

        async def _arunner(
            func: Callable,
            inputs: Union[Dict[str, Any], List[Any], str],
            dataset_item_uuid: str,
            test_uuid: str,
        ):
            # if "_RAW_VALUE_" in inputs:
            if isinstance(inputs, str):
                result = await func(inputs)
            else:
                result = await func(**inputs)

            if not isinstance(result, dict) and not isinstance(result, str):
                result = str(result)

            self._worker.capture_test_observation(
                created_at=datetime.now(timezone.utc),
                name=dataset_name,
                test_uuid=test_uuid,
                dataset_item_uuid=dataset_item_uuid,
                inputs=inputs,
                outputs=result,
            )

        async def run_async(func: Callable, dataset_items: List[DatasetItem]):
            for i in range(0, len(dataset_items), batch_size):
                batch = list(islice(dataset_items, i, i + batch_size))
                coros = [
                    _arunner(func, data.inputs, data.uuid, test_uuid) for data in batch
                ]
                await asyncio.gather(*coros)
                if i + batch_size < len(dataset_items):
                    await asyncio.sleep(delay)

        def run_threaded(func, dataset_items: List[DatasetItem]):
            with ThreadPoolExecutor() as executor:
                for i in range(0, len(dataset_items), batch_size):
                    batch = list(islice(dataset_items, i, i + batch_size))
                    futures = [
                        executor.submit(
                            _runner, func, data.inputs, data.uuid, test_uuid
                        )
                        for data in batch
                    ]
                    for future in futures:
                        future.result()
                    if i + batch_size < len(dataset_items):
                        time.sleep(delay)

        print("Testing...")
        if asyncio.iscoroutinefunction(func):
            asyncio.run(run_async(func, dataset.items))
        else:
            run_threaded(func, dataset.items)

        print("Test finished. Start Logging...")

        self._worker.testing = False
        self.testing = False

        self._worker.flush()
        print("Test completed.")

    def close(self):
        """Close the client connection."""
        self._worker.stop()

    def flush(self):
        """Flush the buffer."""
        self._worker.flush()



from weavel import create_client, WeavelClient, Trace
from uuid import uuid4

client: WeavelClient = create_client()

user_identifier = "USER IDENTIFIER EXAMPLE"

client.identify(user_identifier, {"email": "test@email.com", "name": "testname", "gender": "male"})

client.close()


from typing import Dict, List, Optional, Any
from promptmodel import FunctionModel, init
import os
import json
import asyncio
from asyncio import Queue
from threading import Thread
from concurrent.futures import ThreadPoolExecutor
from supabase import Client, create_client
import logging
import litellm

url = os.getenv("SUPABASE_URL")
key = os.getenv("SUPABASE_KEY")

supabase: Client = create_client(url, key)
init(use_cache=False)

logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("LiteLLM").setLevel(logging.WARNING)

project_uuid = "b7c360b6-a862-4ff3-ad31-d8a620a201dc"

dataset = (
    supabase.table("trace")
    .select("*, trace_data(*)")
    .eq("project_uuid", project_uuid)
    .execute()
).data

for data in dataset:
    # sort trace_data
    data["trace_data"] = sorted(data["trace_data"], key=lambda x: x["created_at"])


# question_extractor = FunctionModel("question_extractor")
answer_extractor = FunctionModel("answer_extractor")
info_extractor = FunctionModel("info_extractor")


# Ensure the directory exists
results_dir = "chat_alpaca_result"
os.makedirs(results_dir, exist_ok=True)

async def worker(queue: Queue):
    while True:
        data = await queue.get()
        result_path = os.path.join(results_dir, f"{data['uuid']}.json")
        try:
            if not os.path.exists(result_path):  # Only process if result doesn't exist
                response = await run_dialog(data)
                res = data.copy()
                res["trace_data"] = response
                # Save result to a file named after the data's uuid
                with open(result_path, "w") as f:
                    f.write(json.dumps(res))
            else:
                print(f"Result for {data['uuid']} already exists. Skipping.")
            queue.task_done()
        except Exception as e:
            print(f"Error processing {data['uuid']}: {e}")
            queue.task_done()

async def run_dialog(data: Dict) -> List[Dict]:
    print(f"START RUN {data['uuid']}")
    # output format: dialog[i]["result"] = {questions: [], answers: {}}
    dialog: Dict[str, Any] = data["trace_data"]
    res = dialog
    # for first message, run question_extractor
    # for subsequent messages, run answer_extractor and run question_extractor
    # try 3 times
    retry_count = 3
    
    retry_count = 3
    while True:
        i_response = await info_extractor.arun_and_parse(
            {
                "current_message" : dialog[0]["content"], 
                "previous_message" : None, 
                "questions": [],
                "previous_topics": []
            }
        )
        # print(i_response)
        if i_response.error is None or i_response.error == False:
            break
    
        if i_response.parsed_outputs == {} or i_response.parsed_outputs is None:
            try:
                parsed_raw_output: Dict = json.loads(i_response.api_response.choices[0].message.content)
                i_response.parsed_outputs = {"informations": parsed_raw_output}
                break
            except Exception as e:
                print(e)
                pass
            
        retry_count -= 1
        if retry_count == 0:
            print("Raise Exception: Failed to run info_extractor")
            raise Exception("Failed to run info_extractor")
    
    # print("-------------------------")
    # print("Extracted Informations")
    # print("-------------------------")
    
    res[0]["question_topics"] = []
    res[0]["information_topics"] = []
     
    res[0]["question_topics"] = [
        {
            "name": q_key,
            "value": q_val["value"],
            "related_to": q_val["related_to"]
        }
        for q_key, q_val in i_response.parsed_outputs["informations"].items() if q_val["type"] == "question"
    ]
    
    res[0]["information_topics"] = [
        {
            "name": info_key,
            "value": info_val["value"],
            "related_to": info_val["related_to"]
        }
        for info_key, info_val in i_response.parsed_outputs["informations"].items() if info_val["type"] == "information"
    ]

    for i in range(1, len(dialog)):
        print(i)
        retry_count = 3
        while True:
            a_response = await answer_extractor.arun_and_parse(
                {
                    "current_message" : dialog[i]["content"],
                    "previous_message" : dialog[i-1]["content"],
                    "questions": [question["name"] for question in dialog[i-1]["question_topics"]]
                }
            )
            # print(a_response)
            if a_response.error is None or a_response.error == False:
                break
            
            if a_response.parsed_outputs == {} or a_response.parsed_outputs is None:
                try:
                    parsed_raw_output = json.loads(a_response.api_response.choices[0].message.content)
                    a_response.parsed_outputs = {"answers": parsed_raw_output}
                    break
                except:
                    pass
                
            retry_count -= 1
            if retry_count == 0:
                print("Raise Exception: Failed to run answer_extractor")
                raise Exception("Failed to run answer_extractor")
        
        # print("-------------------------")
        # print("Extracted Answers")
        # print("-------------------------")
        
        retry_count = 3
        while True:
            i_response = await info_extractor.arun_and_parse(
                {
                    "current_message" : dialog[i]["content"], 
                    "previous_message" : dialog[i-1]["content"], 
                    "questions": [question["name"] for question in dialog[i-1]["question_topics"]],
                    "previous_topics": [info["name"] for info in dialog[i-1]["information_topics"]]
                }
            )
            # print(i_response)
            if i_response.error is None or i_response.error == False:
                break
        
            if i_response.parsed_outputs == {} or i_response.parsed_outputs is None:
                try:
                    parsed_raw_output = json.loads(i_response.api_response.choices[0].message.content)
                    i_response.parsed_outputs = {"informations": parsed_raw_output}
                    break
                except:
                    pass
                
            retry_count -= 1
            if retry_count == 0:
                print("Raise Exception: Failed to run info_extractor")
                raise Exception("Failed to run info_extractor")
            
        # print("-------------------------")
        # print("Extracted Informations")
        # print("-------------------------")

        res[i]["question_topics"] = []
        res[i]["information_topics"] = []
        
        res[i]["question_topics"] = [
            {
                "name": q_key,
                "value": q_val["value"],
                "related_to": q_val["related_to"]
            }
            for q_key, q_val in i_response.parsed_outputs["informations"].items() if q_val["type"] == "question"
        ]
        
        res[i]["information_topics"] = [
            {
                "name": info_key,
                "value": info_val["value"],
                "related_to": info_val["related_to"]
            }
            for info_key, info_val in i_response.parsed_outputs["informations"].items() if info_val["type"] == "information"
        ]
        
        res[i]["information_topics"] += [
            {
                "name": ans_key,
                "value": ans_val,
                "related_to": ans_key
            }
            for ans_key, ans_val in a_response.parsed_outputs["answers"].items()
        ]
        
        
    print(f"END RUN {data['uuid']}")
    return res

async def main():
    queue = Queue()
    num_workers = 30  # Adjust based on your system's capabilities
    tasks = []

    for _ in range(num_workers):
        task = asyncio.create_task(worker(queue))
        tasks.append(task)

    print(len(dataset))
    filtered_dataset = []
    # Pre-execution filtering is now simplified
    for data in dataset:
        result_path = os.path.join(results_dir, f"{data['uuid']}.json")
        if not os.path.exists(result_path):
            filtered_dataset.append(data)
    
    print(len(filtered_dataset))    
    
    await asyncio.sleep(5)
    
    for data in filtered_dataset:
        await queue.put(data)

    await queue.join()
    for task in tasks:
        task.cancel()

    print("All workers are done!")
    
if __name__ == "__main__":

    asyncio.run(main())



"""
Weavel, natural language analysis Dashboard for LLM Agent
"""

from setuptools import setup, find_namespace_packages

# Read README.md for the long description
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="weavel",
    version="1.7.1",
    packages=find_namespace_packages(),
    entry_points={},
    description="Weavel, Prompt Optimization and Evaluation for LLM Applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="weavel",
    url="https://github.com/weavel-ai/weavel-python",
    install_requires=[
        "httpx[http2]",
        "pydantic>=2.4.2",
        "typer[all]",
        "pendulum",
        "requests",
        "cryptography",
        "pyyaml",
        "InquirerPy",
        "python-dotenv",
        "websockets",
        "termcolor",
        "watchdog",
        "readerwriterlock",
        "pendulum",
        "httpx[http2]",
        "nest_asyncio",
    ],
    python_requires=">=3.8.10",
    keywords=[
        "weavel",
        "agent",
        "llm",
        "evaluation",
        "llm evaluation",
        "prompt evaluation",
        "dataset curation",
        "prompt engineering",
        "prompt optimization",
        "AI Prompt Engineer"
    ],
)



from dotenv import load_dotenv
import os
from typing import Dict, Optional
from uuid import uuid4
from openai import AsyncOpenAI, OpenAI
from datetime import datetime, timezone
import time

from weavel._worker import Worker

pricing = {
    "gpt-4o": {"input": 0.000005, "output": 0.000015},
    "gpt-4o-2024-08-06": {"input": 0.0000025, "output": 0.00001},
    "gpt-4o-2024-05-13": {"input": 0.000005, "output": 0.000015},
    "gpt-4o-mini": {"input": 0.00000015, "output": 0.0000006},
    "gpt-4o-mini-2024-07-18": {"input": 0.00000015, "output": 0.0000006},
}

DEFAULT_PARAMS = {
    "model": None,
    "frequency_penalty": 0,
    "logit_bias": None,
    "logprobs": False,
    "top_logprobs": None,
    "max_tokens": None,
    "n": 1,
    "presence_penalty": 0,
    "response_format": None,
    "seed": None,
    "service_tier": None,
    "stop": None,
    "stream": False,
    "stream_options": None,
    "temperature": 1,
    "top_p": 1,
    "tools": None,
    "tool_choice": None,
    "parallel_tool_calls": True,
    "user": None,
}

load_dotenv()

def calculate_cost(model: str, usage: Dict[str, int]) -> float:
    if model not in pricing:
        # raise ValueError(f"Unknown model: {model}")
        return round(0, 6)

    input_tokens = usage.get("prompt_tokens", 0)
    output_tokens = usage.get("completion_tokens", 0)

    input_cost = input_tokens * pricing[model]["input"]
    output_cost = output_tokens * pricing[model]["output"]

    total_cost = input_cost + output_cost

    return round(total_cost, 6)


class LoggingSuppressor:
    def __init__(self, worker):
        self.worker = worker
        self.original_capture_generation = None

    def __enter__(self):
        self.original_capture_generation = self.worker.capture_generation
        self.worker.capture_generation = lambda *args, **kwargs: None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.worker.capture_generation = self.original_capture_generation


class AsyncLoggingSuppressor:
    def __init__(self, worker):
        self.worker = worker
        self.original_capture_generation = None
        self.original_acapture_generation = None

    async def __aenter__(self):
        self.original_capture_generation = self.worker.capture_generation
        self.original_acapture_generation = self.worker.acapture_generation
        self.worker.capture_generation = None
        self.worker.acapture_generation = None

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.worker.capture_generation = self.original_capture_generation
        self.worker.acapture_generation = self.original_acapture_generation


class WeavelOpenAI(OpenAI):
    def __init__(
        self,
        *args,
        weavel_api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.weavel_api_key = weavel_api_key or os.getenv("WEAVEL_API_KEY")
        self.original_chat_completions = self.chat.completions
        self.original_beta_chat_completions = self.beta.chat.completions

        class CustomChatCompletions:
            def __init__(self, original, weavel_api_key: str, base_url: Optional[str]):
                self.original = original
                self.weavel_api_key = weavel_api_key
                self.base_url = base_url
                self._worker = Worker(
                    self.weavel_api_key,
                    self.base_url,
                    flush_interval=3,
                )
                self.header = {}

            def create(self, *args, **kwargs):
                self.header = kwargs.pop("headers", {})
                is_streaming = kwargs.get("stream", False)

                if is_streaming:
                    return self._handle_streaming(*args, **kwargs)
                else:
                    return self._handle_non_streaming(*args, **kwargs)

            def _handle_streaming(self, *args, **kwargs):
                kwargs.setdefault("stream_options", {})
                kwargs["stream_options"]["include_usage"] = True

                start_time = time.time()
                response = self.original.create(*args, **kwargs)
                accumulated_response = {
                    "id": None,
                    "choices": [
                        {
                            "index": 0,
                            "message": {
                                "role": "assistant",
                                "content": "",
                                "function_call": None,
                                "tool_calls": None,
                            },
                            "finish_reason": None,
                        }
                    ],
                    "created": None,
                    "model": None,
                    "object": "chat.completion",
                    "usage": None,
                    "service_tier": None,
                    "system_fingerprint": None,
                }

                for chunk in response:
                    if len(chunk.choices) == 0:
                        response_details = chunk.model_dump()
                        accumulated_response["id"] = response_details["id"]
                        accumulated_response["created"] = response_details["created"]
                        accumulated_response["model"] = response_details["model"]
                        accumulated_response["object"] = response_details["object"]
                        accumulated_response["usage"] = response_details["usage"]
                        accumulated_response["service_tier"] = response_details[
                            "service_tier"
                        ]
                        accumulated_response["system_fingerprint"] = response_details[
                            "system_fingerprint"
                        ]
                        continue
                    else:
                        delta = chunk.choices[0].delta
                        if delta.content:
                            accumulated_response["choices"][0]["message"][
                                "content"
                            ] += delta.content
                        if delta.function_call:
                            if (
                                accumulated_response["choices"][0]["message"][
                                    "function_call"
                                ]
                                is None
                            ):
                                accumulated_response["choices"][0]["message"][
                                    "function_call"
                                ] = {"name": "", "arguments": ""}
                            accumulated_response["choices"][0]["message"][
                                "function_call"
                            ]["name"] += (delta.function_call.name or "")
                            accumulated_response["choices"][0]["message"][
                                "function_call"
                            ]["arguments"] += (delta.function_call.arguments or "")
                        if delta.tool_calls:
                            if (
                                accumulated_response["choices"][0]["message"][
                                    "tool_calls"
                                ]
                                is None
                            ):
                                accumulated_response["choices"][0]["message"][
                                    "tool_calls"
                                ] = []
                            accumulated_response["choices"][0]["message"][
                                "tool_calls"
                            ].extend(delta.tool_calls)
                        if chunk.choices[0].finish_reason:
                            accumulated_response["choices"][0]["finish_reason"] = (
                                chunk.choices[0].finish_reason
                            )

                    yield chunk

                end_time = time.time()
                latency = end_time - start_time
                accumulated_response["latency"] = round(latency, 6)

                model = accumulated_response["model"]
                usage = accumulated_response["usage"]
                cost = calculate_cost(model, usage)
                accumulated_response["cost"] = cost

                self._capture_generation(inputs=kwargs, outputs=accumulated_response)

            def _handle_non_streaming(self, *args, **kwargs):
                start_time = time.time()
                response = self.original.create(*args, **kwargs)
                end_time = time.time()
                latency = end_time - start_time
                latency = round(latency, 6)

                formatted_response = {
                    "id": response.id,
                    "choices": [choice.model_dump() for choice in response.choices],
                    "created": response.created,
                    "model": response.model,
                    "object": response.object,
                    "usage": response.usage.model_dump() if response.usage else None,
                    "service_tier": response.service_tier,
                    "system_fingerprint": response.system_fingerprint,
                    "latency": latency,
                }

                # Calculate and add cost
                cost = calculate_cost(response.model, formatted_response["usage"])
                formatted_response["cost"] = cost

                self._capture_generation(inputs=kwargs, outputs=formatted_response)
                return response

            def _capture_generation(
                self,
                inputs,
                outputs,
            ):
                self._worker.capture_generation(
                    observation_id=str(uuid4()),
                    created_at=datetime.now(timezone.utc),
                    name=self.header.get("name", "OpenAI Chat"),
                    prompt_name=self.header.get("prompt_name", None),
                    inputs=inputs,
                    outputs=outputs,
                )

        class CustomBetaChatCompletions:
            def __init__(self, original, weavel_api_key: str, base_url: Optional[str]):
                self.original = original
                self.weavel_api_key = weavel_api_key
                self.base_url = base_url
                self._worker = Worker(
                    self.weavel_api_key,
                    self.base_url,
                    flush_interval=3,
                )

            def parse(self, *args, **kwargs):
                header = kwargs.pop("headers", {})
                start_time = time.time()

                # context manager to suppress duplicate logging
                with LoggingSuppressor(self._worker):
                    response = self.original.parse(*args, **kwargs)
                    
                end_time = time.time()
                latency = end_time - start_time
                latency = round(latency, 6)
                
                formatted_response = {
                    "id": response.id,
                    "choices": [choice.model_dump() for choice in response.choices],
                    "created": response.created,
                    "model": response.model,
                    "object": response.object,
                    "usage": response.usage.model_dump() if response.usage else None,
                    "service_tier": response.service_tier,
                    "system_fingerprint": response.system_fingerprint,
                    "latency": latency,
                }
                
                # Calculate and add cost
                cost = calculate_cost(response.model, formatted_response["usage"])
                formatted_response["cost"] = cost
                
                self._worker.capture_generation(
                    observation_id=str(uuid4()),
                    created_at=datetime.now(timezone.utc),
                    name=header.get("name", "OpenAI Beta Chat Parse"),
                    prompt_name=header.get("prompt_name", None),
                    inputs=kwargs,
                    outputs=formatted_response,
                )
                
                return response

        self.chat.completions = CustomChatCompletions(
            self.original_chat_completions, self.weavel_api_key, base_url
        )
        self.beta.chat.completions = CustomBetaChatCompletions(
            self.original_beta_chat_completions, self.weavel_api_key, base_url
        )


class AsyncWeavelOpenAI(AsyncOpenAI):
    def __init__(
        self,
        *args,
        weavel_api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.weavel_api_key = weavel_api_key or os.getenv("WEAVEL_API_KEY")
        self.original_chat_completions = self.chat.completions
        self.original_beta_chat_completions = self.beta.chat.completions

        class AsyncCustomChatCompletions:
            def __init__(self, original, weavel_api_key: str, base_url: Optional[str]):
                self.original = original
                self.weavel_api_key = weavel_api_key
                self.base_url = base_url
                self._worker = Worker(
                    self.weavel_api_key,
                    self.base_url,
                    flush_interval=3,
                )
                self.header = {}

            async def create(self, *args, **kwargs):
                self.header = kwargs.pop("headers", {})
                is_streaming = kwargs.get("stream", False)

                if is_streaming:
                    return self._handle_streaming(*args, **kwargs)
                else:
                    return await self._handle_non_streaming(*args, **kwargs)

            async def _handle_streaming(self, *args, **kwargs):
                kwargs.setdefault("stream_options", {})
                kwargs["stream_options"]["include_usage"] = True

                start_time = time.time()
                response = await self.original.create(*args, **kwargs)
                accumulated_response = {
                    "id": None,
                    "choices": [
                        {
                            "index": 0,
                            "message": {
                                "role": "assistant",
                                "content": "",
                                "function_call": None,
                                "tool_calls": None,
                            },
                            "finish_reason": None,
                        }
                    ],
                    "created": None,
                    "model": None,
                    "object": "chat.completion",
                    "usage": None,
                    "service_tier": None,
                    "system_fingerprint": None,
                }

                async for chunk in response:
                    if len(chunk.choices) == 0:
                        response_details = chunk.model_dump()
                        accumulated_response["id"] = response_details["id"]
                        accumulated_response["created"] = response_details["created"]
                        accumulated_response["model"] = response_details["model"]
                        accumulated_response["object"] = response_details["object"]
                        accumulated_response["usage"] = response_details["usage"]
                        accumulated_response["service_tier"] = response_details[
                            "service_tier"
                        ]
                        accumulated_response["system_fingerprint"] = response_details[
                            "system_fingerprint"
                        ]
                        continue
                    else:
                        delta = chunk.choices[0].delta
                        if delta.content:
                            accumulated_response["choices"][0]["message"][
                                "content"
                            ] += delta.content
                        if delta.function_call:
                            if (
                                accumulated_response["choices"][0]["message"][
                                    "function_call"
                                ]
                                is None
                            ):
                                accumulated_response["choices"][0]["message"][
                                    "function_call"
                                ] = {"name": "", "arguments": ""}
                            accumulated_response["choices"][0]["message"][
                                "function_call"
                            ]["name"] += (delta.function_call.name or "")
                            accumulated_response["choices"][0]["message"][
                                "function_call"
                            ]["arguments"] += (delta.function_call.arguments or "")
                        if delta.tool_calls:
                            if (
                                accumulated_response["choices"][0]["message"][
                                    "tool_calls"
                                ]
                                is None
                            ):
                                accumulated_response["choices"][0]["message"][
                                    "tool_calls"
                                ] = []
                            accumulated_response["choices"][0]["message"][
                                "tool_calls"
                            ].extend(delta.tool_calls)
                        if chunk.choices[0].finish_reason:
                            accumulated_response["choices"][0]["finish_reason"] = (
                                chunk.choices[0].finish_reason
                            )

                    yield chunk

                end_time = time.time()
                latency = end_time - start_time
                accumulated_response["latency"] = round(latency, 6)

                model = accumulated_response["model"]
                usage = accumulated_response["usage"]
                cost = calculate_cost(model, usage)
                accumulated_response["cost"] = cost
                
                if self._worker.capture_generation is not None:
                    await self._capture_generation(
                        inputs=kwargs, outputs=accumulated_response
                    )

            async def _handle_non_streaming(self, *args, **kwargs):
                start_time = time.time()
                response = await self.original.create(*args, **kwargs)
                end_time = time.time()
                latency = end_time - start_time
                latency = round(latency, 6)

                formatted_response = {
                    "id": response.id,
                    "choices": [choice.model_dump() for choice in response.choices],
                    "created": response.created,
                    "model": response.model,
                    "object": response.object,
                    "usage": response.usage.model_dump() if response.usage else None,
                    "service_tier": response.service_tier,
                    "system_fingerprint": response.system_fingerprint,
                    "latency": latency,
                }

                # Calculate and add cost
                cost = calculate_cost(response.model, formatted_response["usage"])
                formatted_response["cost"] = cost

                if self._worker.capture_generation is None:
                    return response

                await self._capture_generation(
                    inputs=kwargs, outputs=formatted_response
                )
                return response
            
            async def _capture_generation(
                self,
                inputs,
                outputs,
            ):
                await self._worker.acapture_generation(
                    observation_id=str(uuid4()),
                    created_at=datetime.now(timezone.utc),
                    name=self.header.get("name", "AsyncOpenAI Chat"),
                    prompt_name=self.header.get("prompt_name", None),
                    inputs=inputs,
                    outputs=outputs,
                )

        class AsyncCustomBetaChatCompletions:
            def __init__(self, original, weavel_api_key: str, base_url: Optional[str]):
                self.original = original
                self.weavel_api_key = weavel_api_key
                self.base_url = base_url
                self._worker = Worker(
                    self.weavel_api_key,
                    self.base_url,
                    flush_interval=3,
                )

            async def parse(self, *args, **kwargs):
                header = kwargs.pop("headers", {})
                start_time = time.time()
                
                async with AsyncLoggingSuppressor(self._worker):
                    response = await self.original.parse(*args, **kwargs)
                    
                end_time = time.time()
                latency = end_time - start_time
                latency = round(latency, 6)
                    
                formatted_response = {
                    "id": response.id,
                    "choices": [choice.model_dump() for choice in response.choices],
                    "created": response.created,
                    "model": response.model,
                    "object": response.object,
                    "usage": response.usage.model_dump() if response.usage else None,
                    "service_tier": response.service_tier,
                    "system_fingerprint": response.system_fingerprint,
                    "latency": latency,
                }
                
                # Calculate and add cost
                cost = calculate_cost(response.model, formatted_response["usage"])
                formatted_response["cost"] = cost

                await self._worker.acapture_generation(
                    observation_id=str(uuid4()),
                    created_at=datetime.now(timezone.utc),
                    name=header.get("name", "Async OpenAI Beta Chat Parse"),
                    prompt_name=header.get("prompt_name", None),
                    inputs=kwargs,
                    outputs=formatted_response,
                )
                return response

        self.chat.completions = AsyncCustomChatCompletions(
            self.original_chat_completions, self.weavel_api_key, base_url
        )
        self.beta.chat.completions = AsyncCustomBetaChatCompletions(
            self.original_beta_chat_completions, self.weavel_api_key, base_url
        )
